
/****************************************************************************
*
*	  Module Filename:  Chemistry.cpp
*	  Module Name:      Chemistry
*	  Description:      Defines the network of possible interactions
*                       between Molecule_Types 
*	  Author:           Chris Gordon-Smith
*
*     Copyright (C) 2003, 2005, 2006, 2008 Chris Gordon-Smith
*     http://www.simsoup.info/
*
*     This program is free software; you can redistribute it and/or modify it
*     under the terms of the GNU General Public License Version 2, as published 
*     by the Free Software Foundation.
*
*     This program is distributed in the hope that it will be useful, but
*     WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
*     or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
*     for more details.
*
*     You should have received a copy of the GNU General Public License along
*     with this program; if not, write to the Free Software Foundation, Inc.,
*     59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*
*     See the file Licence.txt for a copy of the GNU General Public License.
*	  Amendment History:-
*
*		Version	   Date	   			Update Description
*		=======    ===========		==================
*
*
*
*
************************************************************************/

//  Comment out to disallow tracing this file
//#define  ENABLE_TRACE_FILE  

#include "globldef.h"
#include "Library_Headers.h"

//--------------  Chemistry Subsystem  --------------------------------------------------------------------------
#include  "Chemistry.h"
#include  "Molecule_Type.h"
#include  "Interaction_Type.h"
//----------------------------------------------------------------------------------------------------------

//---------------- Statistics Subsystem --------------------------------------------------------------------------
#include  "Indexed_Series.h"
#include  "Simulation_Statistics.h"
//----------------------------------------------------------------------------------------------------------------

//  Initialise static variables

Chemistry*  Chemistry::Chemistry__Ptr = NULL;

//------------------------------------------------------------------------------

void    Chemistry::Add_Mol_Type(Molecule_Type*  Mol_Type_Ptr)

//    Add the specified Molecule_Type to the Chemistry

{
    TList__Of_Molecule_Types_Ptr->Add(Mol_Type_Ptr);
    Number_Of_Molecule_Types++;
}

//----------------------------------------------------------------------------------------------------------

void    Chemistry::Add_Interaction_Type(Interaction_Type*  Int_Type_Ptr)

//  Add the specified Interaction_Type to the Chemistry

{
    TList__Of_Interaction_Types_Ptr->Add(Int_Type_Ptr);
}
//----------------------------------------------------------------------------------------------------------

Chemistry::Chemistry()

{
    Chemistry__Ptr = this;
  
//  Create the Lists attached to the Chemistry

    CREATE_OBJECT(VCSList,Chemistry__Ptr->TList__Of_Interaction_Types_Ptr = new VCSList);
    CREATE_OBJECT(VCSList,Chemistry__Ptr->TList__Of_Molecule_Types_Ptr = new VCSList);
    
    Number_Of_Molecule_Types = 0;
    Rate_Const_Model = Arrenhius;  // Default
    
    Initialise();
}
//----------------------------------------------------------------------------------------------------------

Chemistry*  Chemistry::Get_Chemistry__Ptr()

// Return a pointer to the Chemistry object

{     
    return(Chemistry__Ptr);
}
//-----------------------------------------------------------------------------------------------------------------

void Chemistry::Clear()

//  Destroy all the Chemistry data structures, except the Molecule Types

{   

//  Delete all the Interaction_Types
 
    for (bool Continue_Flag = Iter_To_First_Int_Type();
         Continue_Flag == true;
         Continue_Flag = Iter_To_Next_Int_Type())
    BEGIN
        DELETE_OBJECT(Interaction_Type, Get_Current_Int_Type_Ptr()); 
    END   
    
//  Clear the list of Interaction Types
    
    TList__Of_Interaction_Types_Ptr->Clear();                                     
    
//  Clear each Molecule_Type's lists of Interaction_Types
    
    for (bool Continue_Flag = Iter_To_First_Mol_Type();
         Continue_Flag == true;
         Continue_Flag = Iter_To_Next_Mol_Type())
    {
        Get_Current_Mol_Type_Ptr()->Clear_Lists_Of_Interaction_Types();
    }   
    
//  Unlink each Molecule_Type from its Realised Mol Type pointer

    for (bool Continue_Flag = Iter_To_First_Mol_Type();
         Continue_Flag == true;
         Continue_Flag = Iter_To_Next_Mol_Type())
    BEGIN
        Get_Current_Mol_Type_Ptr()->Unlink_From_R_Mol_Type();
    END
    
    Interaction_Type::Clear_Static();
    
    Assert (TList__Of_Interaction_Types_Ptr->Count() == 0, 
        "Chemistry::Clear() - List of Interaction Types is not empty");
        
//  Delete all the Molecule Types

    for (bool Continue_Flag = Iter_To_First_Mol_Type();
         Continue_Flag == true;
         Continue_Flag = Iter_To_Next_Mol_Type())
    BEGIN
        DELETE_OBJECT(Molecule_Type, Get_Current_Mol_Type_Ptr()); 
    END   
    
//  Clear the list of Molecule Types
    
    TList__Of_Molecule_Types_Ptr->Clear();   
    Number_Of_Molecule_Types = 0;     
    Molecule_Type::Reset_Next_Molecule_Type_Identifier();
                                
    Assert (TList__Of_Molecule_Types_Ptr->Count() == 0, 
        "Chemistry::Clear() - List of Molecule Types is not empty");

//  Clear the Chemistry's Indexed Series 
     Num_Constructions_Series_Ptr->Clear();
     Num_Fissions_Series_Ptr->Clear();
     Num_IntTypes_Series_Ptr->Clear();
}
//------------------------------------------------------------------------------

void    Chemistry::Collect_Statistics(int Time )

{
    int Num_Constructions = Get_Number_Of_Constructions();
    Num_Constructions_Series_Ptr->Add_Value(Time, Num_Constructions);

    int Num_Fissions = Get_Number_Of_Fissions();
    Num_Fissions_Series_Ptr->Add_Value(Time, Num_Fissions);

    int Num_Int_Types = Get_Number_Of_Interaction_Types();
    Num_IntTypes_Series_Ptr->Add_Value(Time, Num_Int_Types);
}
//----------------------------------------------------------------------------------------------------------------
Molecule_Type*  Chemistry::Get_MolType_Pt(int MolType_Indx)

//  Get a pointer to the specified Molecule_Type

{
    TRACE("Chemistry::Get_MolType_Pt");
    Assert ((Chemistry__Ptr->TList__Of_Molecule_Types_Ptr->Count() > MolType_Indx)
                    &&
            (-1 < MolType_Indx),
            "Chemistry::Get_MolType_Pt- Invalid Argument");

    Molecule_Type*  MolType_Pt;
    MolType_Pt = static_cast<Molecule_Type*>  
                    (Chemistry__Ptr->TList__Of_Molecule_Types_Ptr->Get_Item(MolType_Indx));

    Assert (MolType_Pt != 0,  "Chemistry::Get_MolType_Pt-  Invalid Mol Type Pointer");
    return (MolType_Pt);
}
//----------------------------------------------------------------------------------------------------------
Interaction_Type*  Chemistry::Get_Current_Int_Type_Ptr() const

//  Get a pointer to the Interaction_Type associated with the current value of the Interaction_Type Iterator


{
    return(static_cast<Interaction_Type*> (TList__Of_Interaction_Types_Ptr->Get_Item(Iter_Interaction_Types)));
}
//----------------------------------------------------------------------------------------------------------

Molecule_Type*  Chemistry::Get_Current_Mol_Type_Ptr() const

//  Get a pointer to the Molecule_Type associated with the current value of the Molecule Type Iterator

{
    return(static_cast<Molecule_Type*> (TList__Of_Molecule_Types_Ptr->Get_Item(Iter_Molecule_Type)));
}
//----------------------------------------------------------------------------------------------------------

int     Chemistry::Get_Num_MolTypes()

//  Return the Number of Molecule Types for the Chemistry

{
    return(Chemistry__Ptr->Number_Of_Molecule_Types);
}
//-----------------------------------------------------------------------------------------------------------------

int Chemistry::Get_Number_Of_Constructions()

//  Return the number of Constructions in the Chemistry

{
    TRACE("Chemistry::Get_Number_Of_Constructions");
    int  Num_Constructions = 0;
    for (bool Continue_Flag = TList__Of_Interaction_Types_Ptr->Iter_To_First();
         Continue_Flag not_eq false;
         Continue_Flag = TList__Of_Interaction_Types_Ptr->Iter_To_Next())
    BEGIN
        Interaction_Type*  IntType_Ptr;
        IntType_Ptr = static_cast<Interaction_Type*> (TList__Of_Interaction_Types_Ptr->Get_Current_Item());
        if (IntType_Ptr->Get_Category() == Interaction_Type_Enumr__Constr)
        THEN
            Num_Constructions++;
        ENDIF
    END
    return Num_Constructions;
}
//----------------------------------------------------------------------------------------------------------------

int Chemistry::Get_Number_Of_Fissions()

//  Return the number of Fissions in the Chemistry

{
    int  Num_Fissions = 0;
    for (bool Continue_Flag = TList__Of_Interaction_Types_Ptr->Iter_To_First();
         Continue_Flag not_eq false;
         Continue_Flag = TList__Of_Interaction_Types_Ptr->Iter_To_Next())
    BEGIN
        Interaction_Type*  IntType_Ptr;
        IntType_Ptr = static_cast<Interaction_Type*> (TList__Of_Interaction_Types_Ptr->Get_Current_Item());
        if (IntType_Ptr->Get_Category() == Interaction_Type_Enumr__Fission)
        THEN
            Num_Fissions++;
        ENDIF
    END
    return Num_Fissions;
}
//----------------------------------------------------------------------------------------------------------------
int  Chemistry::Get_Number_Of_Interaction_Types()

//    Return the number if Interaction Types in the Chemistry

{
    return TList__Of_Interaction_Types_Ptr->Count();
}
//----------------------------------------------------------------------------------------------------------------

Interaction_Type*   Chemistry::Get_Random_IntType_Ptr()

//  Get a pointer to an Interaction Type selected at random

{
    return static_cast<Interaction_Type*> (TList__Of_Interaction_Types_Ptr->Get_Random_Value());
}
//----------------------------------------------------------------------------------------------------------------

Chemistry::Rate_Const_Model_Tp  Chemistry::Get_Rate_Constant_Model()

//  Return the model (paradigm) to be used for detremining Rate Constants in this Chemistry

{
    Assert (Chemistry__Ptr not_eq 0, "Chemistry::Get_Rate_Constant_Model - the Chemistry has not been initialised");
    return Chemistry__Ptr->Rate_Const_Model;
}
//----------------------------------------------------------------------------------------------------------------

void    Chemistry::Initialise()

{
   TRACE("Chemistry::Initialise");
   Simulation_Statistics*  SimStats_Ptr = Simulation_Statistics::Get_Ptr();
                        
   SimStats_Ptr->Create_Indexed_Series_List(string("Chemistry Overview"), true);  // Is for user access

//  Create the Statistics Collection Series

   Num_Constructions_Series_Ptr = SimStats_Ptr->Create_Indexed_Series(string("Constructions"), 
                                                                      string("Time"), 
                                                                      string("Constructions"), 
                                                                      string("Chemistry Overview"),
                                                                      true);

   Num_Fissions_Series_Ptr = SimStats_Ptr->Create_Indexed_Series(string("Fissions"), 
                                                                 string("Time"), 
                                                                 string("Fissions"), 
                                                                 string("Chemistry Overview"),
                                                                 true);

   Num_IntTypes_Series_Ptr = SimStats_Ptr->Create_Indexed_Series(string("Interaction Types"), 
                                                                 string("Time"), 
                                                                 string("Interaction Types"), 
                                                                 string("Chemistry Overview"),
                                                                 true);
}
//----------------------------------------------------------------------------------------------------------------

/* void    Chemistry::Initialise_Chemistry( int UI_Number_Of_Molecule_Types,
                                         int UI_Number_Of_Interaction_Types,
                                         int UI_Maximum_Rate_Constant,
                                         int UI_Percent_Constructions,
                                         int UI_Percent_Fissions,
                                         int UI_Percent_Transformations)
//  Initialise the Chemistry. Setup the specified number of Interaction_Types.
//  Interaction_Types for each Interaction Scheme are setup at random. The
//  number of Interaction_Types of each Interaction Scheme is determined according to the
//  total Number_Of_Interaction_Types.
//  Rate_Constants are also setup.

{    
    int Molecule_Type_Counter;
    Molecule_Type*      Molecule_Type__Ptr;

    int Num_Of_Constructions;
    int Num_Of_Fissions;    
    int Num_Of_Transformations;
    int Construction_Counter;
    int Fission_Counter;   
    int Transformation_Counter;

    //  Set the Parameters

    Chemistry__Ptr->Number_Of_Molecule_Types = UI_Number_Of_Molecule_Types;
    Chemistry__Ptr->Maximum_Rate_Constant = UI_Maximum_Rate_Constant;

//  Now create the number of Molecule_Types requested by the user
//  For Random Chemistry, the Specification is just the sequence number
//  for the Molecule_Type

    for (Molecule_Type_Counter = 0;
         Molecule_Type_Counter < Chemistry__Ptr->Number_Of_Molecule_Types;
         Molecule_Type_Counter++)
    {
        CREATE_OBJECT(Molecule_Type,Molecule_Type__Ptr = new Molecule_Type);

//      Add the Molecule_Type to the Chemistry

        Molecule_Type__Ptr->Add_To_Chemistry();
    }

//  Now create the Interaction_Types as requested by the user

//  First calculate the number of Interaction_Types in each Category

    Num_Of_Constructions = (UI_Percent_Constructions * UI_Number_Of_Interaction_Types)/100;
    Num_Of_Fissions      = (UI_Percent_Fissions * UI_Number_Of_Interaction_Types)/100;
    Num_Of_Transformations  = (UI_Percent_Transformations * UI_Number_Of_Interaction_Types)/100;

//  Setup the Construction Interaction_Types

    for (Construction_Counter = 0;
         Construction_Counter < Num_Of_Constructions;

         Construction_Counter++)
    {
        CREATE_OBJECT(Construction, new Construction);
    }

          
//  Setup the Fission Interaction_Types

    for (Fission_Counter = 0;
         Fission_Counter < Num_Of_Fissions;
         Fission_Counter++)
    {
        CREATE_OBJECT(Fission,new Fission);
    }

//  Setup the Transformation Interaction_Types

    for (Transformation_Counter = 0;
         Transformation_Counter < Num_Of_Transformations;
         Transformation_Counter++)
    {
        CREATE_OBJECT(Transformation, new Transformation);
    }  */
//}
//----------------------------------------------------------------------------------------------------------

bool  Chemistry::Iter_To_First_Int_Type()

//  Move the Iterator to the first Interaction_Type

{
    if(TList__Of_Interaction_Types_Ptr->Count() > 0)
    {
        Iter_Interaction_Types = 0;
        return(true);
    }else{
        return(false);
    }
}
//----------------------------------------------------------------------------------------------------------

bool  Chemistry::Iter_To_First_Mol_Type()

//  Move the Iterator to the first Molecule Type


{
    if(TList__Of_Molecule_Types_Ptr->Count() > 0)
    {
        Iter_Molecule_Type = 0;
        return(true);
    }else{
        return(false);
    }
}

//----------------------------------------------------------------------------------------------------------

bool  Chemistry::Iter_To_Next_Int_Type()

//  Move the Iterator to the next Interaction_Type

{
    if(Iter_Interaction_Types < TList__Of_Interaction_Types_Ptr->Count() - 1)
    {
        Iter_Interaction_Types++;
        return(true);
    }else{

        return(false);
    }
}


//----------------------------------------------------------------------------------------------------------

bool  Chemistry::Iter_To_Next_Mol_Type()

//  Move the Iterator to the next Molecule Type

{
    if(Iter_Molecule_Type < TList__Of_Molecule_Types_Ptr->Count() - 1)
    {
        Iter_Molecule_Type++;
        return(true);
    }else{
        return(false);
    }
}
//----------------------------------------------------------------------------------------------------------------

void  Chemistry::Make()

{
    Assert(Chemistry__Ptr == 0, "Chemistry::Make - Singleton instance pointer is not zero");
    CREATE_OBJECT(Chemistry, Chemistry__Ptr = new Chemistry);
}
//----------------------------------------------------------------------------------------------------------
void    Chemistry::Remove_Interaction_Type(Interaction_Type*  Int_Type_Ptr)

//  Remove the specified Interaction_Type from the Chemistry

{

    TList__Of_Interaction_Types_Ptr->Remove(Int_Type_Ptr);
}
//----------------------------------------------------------------------------------------------------------

//void    Chemistry::Set_Max_Rand_Rate_Constant(int Max_Rate_Const)

//{
//    Maximum_Rand_Rate_Constant = Max_Rate_Const;
//}
//----------------------------------------------------------------------------------------------------------------
void  Chemistry::Set_Number_Of_Molecule_Types(int Num_Mol_Types)

//  Set the Number of Molecule Types for the Chemistry

{
    Number_Of_Molecule_Types = Num_Mol_Types;
}
//-----------------------------------------------------------------------------------------------------------------

bool Chemistry::Set_Rate_Contant_Model(Rate_Const_Model_Tp  Model)

//  Set the Model to be used for determining Rate Constants
//  Return false (error) if it has already been set

{
    Assert(Chemistry__Ptr not_eq 0, 
	   "Chemistry::Set_Rate_Contant_Model - Attempt to set Rate Constant Model before Chemistry setup");

    if (Chemistry__Ptr->Rate_Const_Model not_eq None)
    THEN
	return false;
    ELSE
	Chemistry__Ptr->Rate_Const_Model = Model;
        return true;
    ENDIF
}
